/* This file name is rt.cpp */
/* C Program for Finding the Roots */
/*   of Real-Coefficient Polynomials */
/* IEEE Trans. on Ed., Vol 37, No.1, p.46, Feb 1994 */
/* Written by Hyung W. Paik */
/* This program is written for Turbo C */

/* If you have any suggestions or comments, */
/* please feel free to drop me a e-mail note. */
/* e-mail: hpaik@silver.sdsmt.edu */

/* Disclaimer */
/* I do not warrant this program for correctness, */
/* accuracy. You rely on the results of the program */
/* solely at your own risk. I disclaims all liability */
/* incurred by this program and third-party claims */
/* against you for losses or damages. */

# define Greal long double
typedef struct COMPLEX { Greal r,i ; } complex;
# define Maxorder 50
# define Maxiter 70
Greal Errstp=1.e-15;
# include <stdio.h>
# include <stdlib.h>
# include <math.h>
# include <time.h>
void initial(complex a[],complex x[],int K,int *MM);
void find(complex fn[],complex *zero,int degree,
	  Greal error[]);
void oneroot(complex *fnreduce,complex *xsol,int k,
	     Greal *err);
complex Cadd(complex a,complex b);
complex Csub(complex a,complex b);
complex Cmul(complex a,complex b);
complex Cdiv(complex a,complex b);
complex Complex(Greal a,Greal b);
Greal Cabs(complex z);
complex Conjg(complex z);
complex Csqrt(complex z);
complex RCmul(Greal x,complex a);
void main() {
     complex fn[Maxorder],zero[Maxorder];
     int n,degree=0,j,degreef=0,degreep=0;
     clock_t tstart,tend;
     Greal error[Maxorder],a[Maxorder],t;
     for (n=0;n<Maxorder;n++) a[n]=fn[n].r=fn[n].i=0.;
     printf("\n This program finds the roots of the "
	    "real\n coefficients polynominal which "
	    "is composed of\n the factored form and "
	    "polynominal form.\n\n"
	    "     Written by H.W. PAIK\n\n");
     printf("Enter the order of the (x-a(1))..."
	    "..(x-a(degree))\n");
     scanf("%d",&degreef);
     a[0]=0.;
     if (degreef>=1) {
	 for (n=1;n<=degreef;n++) {
	      printf("  Enter  real a(%d)\n   ",n);
	      scanf("%Lg",&a[n]);
	 }
     }
     printf("\nEnter the degree of non-factored "
	    "adding polynominal\n");
     scanf("%d",&degreep);
     if (degreep >=0) {
	 for (n=degreep;n>=0;n--) {
	      printf("   Enter the coefficient of"
		     " x^%d\n   ",n);
	      scanf("%Lg",&fn[n].r);
	 }      }
     if ( (degreef+degreep) == 0 ) return;
     degree=degreep;
     if ( degreef > degreep ) degree=degreef;
     if (degreef >= 1) {
	 for (n=1;n<=degreef;n++) {
	    if (n==1) {a[0]=-a[n];a[1]=1.;continue;}
	    t=a[n]; a[n]=a[n-1];
	    for (j=n-1;j>=1;j--) a[j]=a[j-1]-a[j]*t;
	    a[0]*=(-t);
	 }
     }
     for (n=0;n<=degree;n++) fn[n].r+=a[n];
     printf("\nReal coefficients of polynominal"
	    " are\n");
     for (n=degree;n>=0;n--)
	 printf("    x^%d  %20.16Lg\n",n,fn[n].r);
     tstart=clock();
     find(fn,zero,degree,error);
     tend=clock();
     for(n=1;n<=degree;n++)
	 printf("%d th root is %18.15Lg j %18.15Lg"
		" error= %10.5Lg\n"
		,n,zero[n].r,zero[n].i,error[n]);
     printf("time used   %f\n\n",
	     ((tend-tstart)/CLK_TCK));
     return;
} /* main */

void find(complex fn[],complex *zero,int degree,
	  Greal error[]) {
     int n,m,mm,MM,p;
     int toggle=0;
     complex fnreduce[Maxorder],xin[Maxorder];
     complex xsol,a,b;
     Greal err;
     for (n=0;n<=degree;n++) fnreduce[n]=fn[n];
     initial(fnreduce,xin,degree,&MM);
     mm=0;
     for (n=degree;n>=1;n--) {
	 if (MM == -1) xsol=xin[0];
	 else if (toggle==0 && mm<=MM) { 
	      xsol=xin[mm];mm++;
	 }
	 else if (toggle==1) xsol.i=-xsol.i;
	 else xsol=Complex(0.,0.);
	 oneroot(fnreduce,&xsol,n,&err);
	 if ( fabs(xsol.i) <= Errstp*fabs(xsol.r) )
	    xsol.i=0.;
	 if ( fabs(xsol.r) <= Errstp*fabs(xsol.i) ) 
	    xsol.r=0.;
	 zero[n]=xsol;
	 a=fnreduce[n];
	 for (m=n-1;m>=0;m--) {
	     b=fnreduce[m]; 	     fnreduce[m]=a;
	     a=Cadd(Cmul(xsol,a),b);
	 }   
	 if (xsol.i != 0. && toggle==0) toggle=1;
	 else toggle=0;
     }
     for (n=degree;n>=1;n--)
	  oneroot(fn,&zero[n],degree,&error[n]);
     for (n=2;n<=degree;n++) {
	  xsol=zero[n];
	  err=error[n];
	  for (m=n-1;m>=1;m--) {
	       if (zero[m].r <= xsol.r ) break;
	       zero[m+1]=zero[m];
	       error[m+1]=error[m];
	  }
	  zero[m+1]=xsol;error[m+1]=err;
     }
} /* find */

void initial(complex a[],complex xin[],
	     int k,int *m) {
     Greal y,y_a,yold,yold_a,xi,xf,xd,xold,x;
     int i,j,N=6,M=6,n=0;
     *m=0;
     for (i=1;i<=3;i++) {
	if (i==1) { 
	   xi=-pow10(N);
	   xf=-pow10(-M);
	   xd=0.1;
	}
	if (i==2) {xi=0.;xf=0.;xd=0.;}
	if (i==3) {xi=pow10(-M);xf=pow10(N);xd=10.;}
	for (x=xi;x<=xf;x*=xd) {
	   y=a[k].r;
	   for (j=k-1;j>=0;j--)  y=x*y+a[j].r;
	   y_a=fabs(y);
	n++;
	if ( y==0.) { 
	   xin[*m].r=x;xin[*m].i=0.;(*m)++;
	}
	if (n==1) { yold=y;yold_a=y_a;xold=x; }
	else {
	   if ( (y*yold) < 0.) {
	      xin[*m].r=x;xin[*m].i=0.;(*m)++;
	   }
	   else if (y_a <= yold_a) {
	      yold_a=y_a;xold=x;
	   }
	   yold=y;
	}
	if (i==2) break;
        } /* for x */
     } /* for i */
     if (*m == 0) { xin[0].r=xold;xin[0].i=0.; }      else (*m)--;
} /* initial */

void oneroot(complex fn[],complex *x,int k,
	     Greal *y_a) {
   int j,iter,last,try=0,again;
   Greal dx_a,x_a;
   Greal gp_a,gm_a;
   complex sq,h,gp,gm,gsq,g,y,yd,ydd_h,dx;
   Greal ysave_a=1.e200,errstp;
   complex xsave=*x;
   for (again=1;again<=20;again++) {
     if(try != 0) {
	 if ( xsave.r != 0 )
	    x->r = 10.*(1.-rand()/16384.)*xsave.r;
	 else x->r = 1.-rand()/16384.;
	 if ( xsave.i != 0 )
	    x->i = 10.*(1.-rand()/16384.)*xsave.i;
	 else x->i = 1.-rand()/16384.;
	 try=0;
     }
     for (iter=1,last=1;iter<=Maxiter;iter++) {
	 y=fn[k];
	 yd=ydd_h=Complex(0.,0.);
	 x_a=Cabs(*x);
	 for (j=k-1;j>=0;j--) {
	     ydd_h=Cadd(Cmul(*x,ydd_h),yd);
	     yd=Cadd(Cmul(*x,yd),y);
	     y=Cadd(Cmul(*x,y),fn[j]);
	 }
	 *y_a=Cabs(y);
	 if (*y_a < ysave_a) {
	     xsave=*x;ysave_a=*y_a;last=iter;
	 }
	 errstp=Errstp*Cabs(fn[0]);
	 if (errstp==0.) errstp=Errstp;
	 if ( *y_a <= errstp ) return;
	 if ( iter-last==7 ) { try=2;break; }
	 g=Cdiv(yd,y); gsq=Cmul(g,g);
	 h=Csub(gsq,RCmul(2.,Cdiv(ydd_h,y)));
	 sq=Csqrt(RCmul(k-1,Csub(RCmul(k,h),gsq)));
	 gp=Cadd(g,sq); gp_a=Cabs(gp);
	 gm=Csub(g,sq); gm_a=Cabs(gm);
	 if ( gp_a < gm_a ) { gp=gm; gp_a=gm_a; }
	 if ( gp_a == 0. ) { try=1;break; }
	 else  dx=Cdiv(Complex(k,0.),gp);
	 dx_a=Cabs(dx);
	 if ( x_a !=0 && (dx_a/x_a) >=1.e6 )
	    dx=RCmul(1.e6/dx_a*x_a,dx);
	 if (dx_a < Errstp*x_a) return;
	 *x=Csub(*x,dx);
     } /* iter */
     *x=xsave;*y_a=ysave_a;
   } /* again */
} /* oneroot */ 
complex Cadd(complex a,complex b) {
    complex c;
    c.r=a.r+b.r; c.i=a.i+b.i;
    return c;
} /* Cadd */

complex Csub(complex a,complex b) {
    complex c;
    c.r=a.r-b.r; c.i=a.i-b.i;
    return c;
} /* Csub */
				    
complex Cmul(complex a,complex b) { 
    complex c;
    c.r=a.r*b.r-a.i*b.i; c.i=a.r*b.i+a.i*b.r;
    return c;
} /* Cmul */

complex Cdiv(complex a,complex b) {
   complex c; Greal r,den;
     if(fabs(b.r) >= fabs(b.i)) {
	  r=b.i/b.r; den=b.r+r*b.i;
	  c.r=(a.r+r*a.i)/den;
	  c.i=(a.i-r*a.r)/den;
     } else {
	  r=b.r/b.i; den=b.i+r*b.r;
	  c.r=(a.r*r+a.i)/den;
	  c.i=(a.i*r-a.r)/den;
     }
   return c;
} /* Cdiv */

complex Complex(Greal a,Greal b) {
   complex c;
   c.r=a; c.i=b;
   return c;
} /* Ccomplex */

Greal Cabs(complex z) {
  Greal x,y,a,t;
  x=fabs(z.r); y=fabs(z.i);
  if (x==0.) a=y;
  else if(y==0.) a=x;
  else if(x > y) {
     t=y/x;a=x*sqrt(1.+t*t);
  } else {
     t=x/y;a=y*sqrt(1.+t*t);
  }
  return a;
} /* Cabs */

complex Conjg(complex z) {
   complex c;
   c.r=z.r; c.i=-z.i;    return c;
} /* Conjg */

complex Csqrt(complex z) {
     Greal x,y,w,r; complex c;
     if ((z.r==0.) && (z.i==0.)) {
	 c.r=c.i=0.; return c;
     } else {
	 x=fabs(z.r);y=fabs(z.i);
	 if(x>=y) {
	      r=y/x;w=sqrt(x)*sqrt(.5*(1.+sqrt(1.+r*r)));
	 } else {
	      r=x/y;w=sqrt(y)*sqrt(.5*(r+sqrt(1.+r*r)));
	 }
	 if (z.r >=0.) { c.r=w;c.i=z.i/(2.*w);
	 } else { c.i=(z.i >=0.) ? w: -w;c.r=z.i/(2.*c.i); }
	 return c;
     }
} /* Csqrt */

complex RCmul(Greal x,complex a) {
   complex c;
   c.r=x*a.r; c.i=x*a.i;
   return c;
} /* RCmul */
